/****************************************************************************
*
*   Copyright (c) 2012 Pau Pajuelo
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License version 2 as
*   published by the Free Software Foundation.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
*   MA 02110-1301 USA.
*
****************************************************************************
*
*   Decode EDID information via I2C Bus (Using first 128 bytes EDID Standard)
*
****************************************************************************
*
*   This program have code from:
*   Edid-decode: Adam Jackson <ajax@nwnk.net>
*   i2cget.c:    Jean Delvare <khali@linux-fr.org>
*   read-edid:   http://www.polypux.org/
*
****************************************************************************/

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "i2c-dev.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <stdint.h>

//1==enabled 0==disable
#define FORCE 1

//Address DVI Standard: 0x50
#define ADDRESS 0x50

const struct {  int x, y, refresh;}
established_timings[] =
{
  /* 0x23 bit 7 - 0 */
  {720, 400, 70},
  {720, 400, 88},
  {640, 480, 60},
  {640, 480, 67},
  {640, 480, 72},
  {640, 480, 75},
  {800, 600, 56},
  {800, 600, 60},
  /* 0x24 bit 7 - 0 */
  {800, 600, 72},
  {800, 600, 75},
  {832, 624, 75},
  {1280, 768, 87},
  {1024, 768, 60},
  {1024, 768, 70},
  {1024, 768, 75},
  {1280, 1024, 75},
  /* 0x25 bit 7*/
  {1152, 870, 75},
  //XXX todo: implement manufacture reserved timings
};

//Parse I2C bus
//Return I2CBus
int lookup_i2c_bus(const char *i2cbus_arg)
{
	unsigned long i2cbus;
	char *end;

	i2cbus = strtoul(i2cbus_arg, &end, 0);
	if (i2cbus > 0xFFFFF)
	{
		return -1;
	}
	return i2cbus;
}

//Open i2c device
//Return file descriptor
//Error<0 and Correct==address
int open_i2c_dev(int i2cbus, char *filename, size_t size, int quiet)
{
	int file;

	snprintf(filename, size, "/dev/i2c/%d", i2cbus);
	filename[size - 1] = '\0';
	file = open(filename, O_RDWR);

	if (file < 0 && (errno == ENOENT || errno == ENOTDIR))
	{
		sprintf(filename, "/dev/i2c-%d", i2cbus);
		file = open(filename, O_RDWR);
	}

	if (file < 0 && !quiet)
	{
		if (errno == ENOENT)
		{
			fprintf(stderr, "Error: Could not open file "
				"`/dev/i2c-%d' or `/dev/i2c/%d': %s\n",
				i2cbus, i2cbus, strerror(ENOENT));
		}
		else
		{
			fprintf(stderr, "Error: Could not open file "
				"`%s': %s\n", filename, strerror(errno));
			if (errno == EACCES)
				fprintf(stderr, "Run as root?\n");
		}
	}
	return file;
}

//With force, let the user read from/write to the registers even when a driver is also running
//Error<0 Correct=0
int set_slave_addr(int file, int address, int force)
{
	if (ioctl(file, force ? I2C_SLAVE_FORCE : I2C_SLAVE, address) < 0)
	{
		return -1;
	}
	return 0;
}

//Decode information about display manufacture
//Return manufacture information
char *manufacturer_name(unsigned char *x)
{
    static char name[4];
    int manufacturer_name_well_formed = 0;

    name[0] = ((x[0] & 0x7C) >> 2) + '@';
    name[1] = ((x[0] & 0x03) << 3) + ((x[1] & 0xE0) >> 5) + '@';
    name[2] = (x[1] & 0x1F) + '@';
    name[3] = 0;

    if (isupper(name[0]) && isupper(name[1]) && isupper(name[2]))
        manufacturer_name_well_formed = 1;

    return name;
}

//Decode detailed information obout display
//1==Valid Data 0=Error
int detailed_block(unsigned char *x, int in_extension,int *did_detailed_timing,int *claims_one_point_four)
{
    static unsigned char name[53];
    int ha, hbl, hso, hspw, hborder, va, vbl, vso, vspw, vborder;
    int i;
    char phsync, pvsync, *syncmethod;

    int has_valid_descriptor_pad = 1,has_valid_range_descriptor = 1,has_valid_max_dotclock = 1;
    int manufacturer_name_well_formed = 0,seen_non_detailed_descriptor = 0,has_valid_dummy_block = 1;
    int has_valid_descriptor_ordering = 1,warning_zero_preferred_refresh = 0,has_valid_cvt = 1;
    int has_name_descriptor = 0,name_descriptor_terminated = 0,has_range_descriptor = 0;
    int warning_excessive_dotclock_correction = 0;

    #if 0
    printf("Hex of detail: ");
    for (i = 0; i < 18; i++)
        printf("%02x", x[i]);
    printf("\n");
    #endif

    if (x[0] == 0 && x[1] == 0)
    {
        //Monitor descriptor block, not detailed timing descriptor.
        if (x[2] != 0)
        {
            //1.3, 3.10.3
            printf("Monitor descriptor block has byte 2 nonzero (0x%02x)\n",x[2]);
            has_valid_descriptor_pad = 0;
        }
        if (x[3] != 0xfd && x[4] != 0x00)
        {
            //1.3, 3.10.3
            printf("Monitor descriptor block has byte 4 nonzero (0x%02x)\n",x[4]);
            has_valid_descriptor_pad = 0;
        }

        seen_non_detailed_descriptor = 1;
        if (x[3] <= 0xF)
        {
                 //in principle we can decode these, if we know what they are.
                 //0x0f seems to be common in laptop panels.
                 //0x0e is used by EPI: http://www.epi-standard.org/

            printf("\nManufacturer-specified data, tag %d\n", x[3]);
            return 1;
        }
        switch (x[3])
        {
            case 0x10:
                printf("\nDummy block\n");
                for (i = 5; i < 18; i++)
                    if (x[i] != 0x00)
                        has_valid_dummy_block = 0;
                return 1;
            case 0xF7:
                /* TODO */
                printf("\nEstablished timings III\n");
                return 1;
            case 0xF8:
            {
                int valid_cvt = 1; // just this block
                printf("CVT 3-byte code descriptor:\n");
                if (x[5] != 0x01)
                {
                    has_valid_cvt = 0;
                    return 0;
                }
                for (i = 0; i < 4; i++)
                    valid_cvt &= detailed_cvt_descriptor(x + 6 + (i * 3), (i == 0));
                has_valid_cvt &= valid_cvt;
                return valid_cvt;
            }
            case 0xF9:
                /* TODO */
                printf("\nColor management data\n");
                return 1;
            case 0xFA:
                /* TODO */
                printf("\nMore standard timings\n");
                return 1;
            case 0xFB:
                /* TODO */
                printf("\nColor point\n");
                return 1;
            case 0xFC:
                //XXX should check for spaces after the \n
                //XXX check: terminated with 0x0A, padded with 0x20
                has_name_descriptor = 1;
                if (strchr((char *)name, '\n'))
                    return 1;
                strncat((char *)name, (char *)x + 5, 13);
                if (strchr((char *)name, '\n'))
                {
                    name_descriptor_terminated = 1;
                    printf("\nMonitor name: %s", name);
                }
                return 1;
            case 0xFD:
            {
                int h_max_offset = 0, h_min_offset = 0;
                int v_max_offset = 0, v_min_offset = 0;
                int is_cvt = 0;
                has_range_descriptor = 1;

                 //XXX todo: implement feature flags, vtd blocks
                 //XXX check: ranges are well-formed; block termination if no vtd

                if (claims_one_point_four)
                {
                    if (x[4] & 0x02)
                    {
                        v_max_offset = 255;
                        if (x[4] & 0x01)
                        {
                            v_min_offset = 255;
                        }
                    }
                    if (x[4] & 0x04)
                    {
                        h_max_offset = 255;
                        if (x[4] & 0x03)
                        {
                            h_min_offset = 255;
                        }
                    }
                }
                else if (x[4])
                {
                    has_valid_range_descriptor = 0;
                }

                 //despite the values, this is not a bitfield.
                 //XXX only 0x00 and 0x02 are legal for pre-1.4

                switch (x[10])
                {
                    case 0x00: //default gtf
                        break;
                    case 0x01: //range limits only
                        break;
                    case 0x02: //secondary gtf curve
                        break;
                    case 0x04: //cvt
                        is_cvt = 1;
                        break;
                    default: //invalid
                        break;
                }

                if (x[5] + v_min_offset > x[6] + v_max_offset)
                    has_valid_range_descriptor = 0;
                if (x[7] + h_min_offset > x[8] + h_max_offset)
                    has_valid_range_descriptor = 0;
                printf("Monitor ranges: %d-%dHZ vertical, %d-%dkHz horizontal",
                    x[5] + v_min_offset, x[6] + v_max_offset,
                    x[7] + h_min_offset, x[8] + h_max_offset);
                if (x[9])
                    printf(", max dotclock %dMHz\n", x[9] * 10);
                else
                {
                    if (claims_one_point_four)
                        has_valid_max_dotclock = 0;
                    printf("\n");
                }

                if (is_cvt)
                {
                    int max_h_pixels = 0;

                    printf("CVT version %d.%d\n", x[11] & 0xf0 >> 4, x[11] & 0x0f);

                    if (x[12] & 0xfc)
                    {
                        int raw_offset = (x[12] & 0xfc) >> 2;
                        printf("Real max dotclock: %.2fMHz\n",(x[9] * 10) - (raw_offset * 0.25));
                        if (raw_offset >= 40)
                            warning_excessive_dotclock_correction = 1;
                    }

                    max_h_pixels = x[12] & 0x03;
                    max_h_pixels <<= 8;
                    max_h_pixels |= x[13];
                    max_h_pixels *= 8;
                    if (max_h_pixels)
                        printf("Max active pixels per line: %d\n", max_h_pixels);

                    printf("Supported aspect ratios: %s %s %s %s %s\n",
                        x[14] & 0x80 ? "4:3" : "",
                        x[14] & 0x40 ? "16:9" : "",
                        x[14] & 0x20 ? "16:10" : "",
                        x[14] & 0x10 ? "5:4" : "",
                        x[14] & 0x08 ? "15:9" : "");
                    if (x[14] & 0x07)
                        has_valid_range_descriptor = 0;

                    printf("Preferred aspect ratio: ");
                    switch((x[15] & 0xe0) >> 5)
                    {
                        case 0x00: printf("4:3"); break;
                        case 0x01: printf("16:9"); break;
                        case 0x02: printf("16:10"); break;
                        case 0x03: printf("5:4"); break;
                        case 0x04: printf("15:9"); break;
                        default: printf("(broken)"); break;
                    }
                    printf("\n");

                    if (x[15] & 0x04)
                        printf("Supports CVT standard blanking\n");
                    if (x[15] & 0x10)
                        printf("Supports CVT reduced blanking\n");

                    if (x[15] & 0x07)
                        has_valid_range_descriptor = 0;

                    if (x[16] & 0xf0)
                    {
                        printf("Supported display scaling:\n");
                        if (x[16] & 0x80)
                            printf("    Horizontal shrink\n");
                        if (x[16] & 0x40)
                            printf("    Horizontal stretch\n");
                        if (x[16] & 0x20)
                            printf("    Vertical shrink\n");
                        if (x[16] & 0x10)
                            printf("    Vertical stretch\n");
                    }

                    if (x[16] & 0x0f)
                        has_valid_range_descriptor = 0;

                    if (x[17])
                        printf("Preferred vertical refresh: %d Hz\n", x[17]);
                    else
                        warning_zero_preferred_refresh = 1;
                }

                 //Slightly weird to return a global, but I've never seen any
                 //EDID block wth two range descriptors, so it's harmless.

                return has_valid_range_descriptor;
            }
            case 0xFE:

                //TODO: Two of these in a row, in the third and fouth slots,
                //seems to be specified by SPWG: http://www.spwg.org/

                // XXX check: terminated with 0x0A, padded with 0x20
                printf("\nASCII string: %s", x+5);
                return 1;
            case 0xFF:
                // XXX check: terminated with 0x0A, padded with 0x20
                printf("\nSerial number: %s", x+5);
                return 1;
            default:
                printf("\nUnknown monitor description type %d\n", x[3]);
                return 0;
        }
    }

    if (seen_non_detailed_descriptor && !in_extension)
    {
        has_valid_descriptor_ordering = 0;
    }
    (*did_detailed_timing) = 1;
    ha = (x[2] + ((x[4] & 0xF0) << 4));
    hbl = (x[3] + ((x[4] & 0x0F) << 8));
    hso = (x[8] + ((x[11] & 0xC0) << 2));
    hspw = (x[9] + ((x[11] & 0x30) << 4));
    hborder = x[15];
    va = (x[5] + ((x[7] & 0xF0) << 4));
    vbl = (x[6] + ((x[7] & 0x0F) << 8));
    vso = ((x[10] >> 4) + ((x[11] & 0x0C) << 2));
    vspw = ((x[10] & 0x0F) + ((x[11] & 0x03) << 4));
    vborder = x[16];
    switch ((x[17] & 0x18) >> 3)
    {
        case 0x00:
            syncmethod = " analog composite";
            break;
        case 0x01:
            syncmethod = " bipolar analog composite";
            break;
        case 0x02:
            syncmethod = " digital composite";
            break;
        case 0x03:
            syncmethod = "";
            break;
    }
    pvsync = (x[17] & (1 << 2)) ? '+' : '-';
    phsync = (x[17] & (1 << 1)) ? '+' : '-';

    printf("\nDetailed mode:\tClock:\t\t\t\t%.3f MHz,\n"
	   "\t\tScreen:\t\t\t\t%d mm x %d mm,\n"
	   "\t\tActive resolution:\t\t%dx%d,\n"
	   "\t\tBlanking resolution:\t\t%dx%d,\n"
	   "\t\tSync offset resolution:\t\t%dx%d,\n"
	   "\t\tSync pulse width resolution:\t%dx%d,\n"
	   "\t\tHborder: %d, Vborder: %d,\n"
	   "\t\t%chsync %cvsync%s%s\n\n",
	    (x[0] + (x[1] << 8)) / 100.0,
	    (x[12] + ((x[14] & 0xF0) << 4)),
	    (x[13] + ((x[14] & 0x0F) << 8)),
	   ha,va, ha + hso,va + vso,ha + hso + hspw,va + vso + vspw,
	   ha + hbl, va + vbl,hborder,vborder,
	   phsync, pvsync, syncmethod, x[17] & 0x80 ? " interlaced" : ""
	  );
    // XXX flag decode
    return 1;
}

//Comprove CVT
//Valid=1 Not valid=0
int detailed_cvt_descriptor(unsigned char *x, int first)
{
    const unsigned char empty[3] = { 0, 0, 0 };
    char *names[] = { "50", "60", "75", "85" };
    int width, height;
    int valid = 1;
    int fifty = 0, sixty = 0, seventyfive = 0, eightyfive = 0, reduced = 0;

    if (!first && !memcmp(x, empty, 3))
        return valid;

    height = x[0];
    height |= (x[1] & 0xf0) << 4;
    height++;
    height *= 2;

    switch (x[1] & 0x0c)
    {
    case 0x00:
        width = (height * 4) / 3; break;
        case 0x04:
        width = (height * 16) / 9; break;
        case 0x08:
        width = (height * 16) / 10; break;
        case 0x0c:
        width = (height * 15) / 9; break;
    }

    if (x[1] & 0x03)
        valid = 0;
    if (x[2] & 0x80)
        valid = 0;
    if (!(x[2] & 0x1f))
        valid = 0;

    fifty	= (x[2] & 0x10);
    sixty	= (x[2] & 0x08);
    seventyfive = (x[2] & 0x04);
    eightyfive  = (x[2] & 0x02);
    reduced	= (x[2] & 0x01);

    if (!valid)
    {
        printf("    (broken)\n");
    }
    else
    {
        printf("    %dx%d @ ( %s%s%s%s%s) Hz (%s%s preferred)\n", width, height,
            fifty ? "50 " : "",
            sixty ? "60 " : "",
            seventyfive ? "75 " : "",
            eightyfive ? "85 " : "",
            reduced ? "60RB " : "",
            names[(x[2] & 0x60) >> 5],
            (((x[2] & 0x60) == 0x20) && reduced) ? "RB" : "");
    }

    return valid;
}

//MAIN PROGRAM
//Error<0 and Correct==0
int main (int argc, char *argv[])
{
    int i2cbus,fd,result, edidint[256], i,len,opt,analog;
    char device[30]="",filename[20]="";
    unsigned char edidchar[257]="";
    time_t the_time;
    struct tm *ptm;

    int has_valid_week = 0,has_valid_year = 0,claims_one_point_oh = 0,did_detailed_timing = 0;
    int claims_one_point_two = 0,claims_one_point_three = 0,claims_one_point_four = 0;
    int nonconformant_digital_display = 0,has_preferred_timing = 0,has_valid_detailed_blocks = 0;

    //Getopt part:
    while ((opt = getopt(argc, argv, "d:")) != -1)//d=device
        switch (opt)
        {
            case 'd':
            sprintf(device,"%s",optarg);
            break;
            default: // ’?’
            fprintf(stderr, "Usage: %s \n",argv[0]);
            return -1;
        }

    if (optind < argc)
    {
        printf("non-option ARGV-elements: ");
        while (optind < argc)
            printf("%s ", argv[optind++]);
        printf("\n");
    }

    if (device[0]!='\0' )
    {
        //MAIN CODE
        i2cbus=lookup_i2c_bus(device);
        if (i2cbus<0)
        {
            fprintf(stderr,"");
            return -1;
        }

        if (ADDRESS < 0x03 || ADDRESS > 0x77)
        {
            printf("Error: Chip address out of range (0x03-0x77)!\n");
            return -1;
        }
        if (ADDRESS < 0)
        {
            fprintf(stderr,"");
            return -1;
        }

        fd=open_i2c_dev(i2cbus, filename, sizeof(filename),0);

        if (fd<0 || set_slave_addr(fd, ADDRESS,FORCE))
        {
            printf("Error: Could not set address to 0x%02x\n",ADDRESS);
            return -1;
        }

        for (i=0;i<=255;i++)
        {
            edidint[i]=i2c_smbus_read_byte_data(fd,i);
            if (edidint[i]<0)
            {
                printf("Error:read failed\n");
                return -1;
            }
        }
        close (fd);

        //Comprove header
        if (edidint[0]==0x00&&edidint[7]==0x00&&edidint[1]==0xff&&edidint[2]==0xff&&edidint[3]==0xff&&edidint[4]==0xff&&edidint[5]==0xff&&edidint[6]==0xff)
        {
            if (edidint[128] == 0xff)
                len = 128;
            else
                len = 256;

            printf("%i-byte EDID successfully retrieved from i2c bus %i\nDecode EDID information:\n\n", len, i2cbus);

            //Undecoded EDID
            #if 0
            for (i=0;i<len;i++)
            {
                printf("%2x-", edidint[i]);
            }
            printf("\n\n");
            #endif

            //Convert integer to unsigned char
            for (i=0;i<=(len-1);i++)
            {
                edidchar[i]=(unsigned char)edidint[i];
            }
            edidint[len]='\0';

            //Decoding part:
            //Manufacture part:
            printf("Manufacturer: %s Model %x Serial Number %u\n",
            manufacturer_name(edidchar + 0x08),
            (unsigned short)(edidchar[0x0A] + (edidchar[0x0B] << 8)),
            (unsigned int)(edidchar[0x0C] + (edidchar[0x0D] << 8)
                   + (edidchar[0x0E] << 16) + (edidchar[0x0F] << 24)));
            // XXX need manufacturer ID table

            //Time part:
            time(&the_time);
            ptm = localtime(&the_time);
            if (edidchar[0x10] < 55 || edidchar[0x10] == 0xff)
            {
                has_valid_week = 1;
                if (edidchar[0x11] > 0x0f)
                {
                    if (edidchar[0x10] == 0xff)
                    {
                        has_valid_year = 1;
                        printf("Made week %hd of model year %hd\n", edidchar[0x10], edidchar[0x11]);
                    }
                    else if (edidchar[0x11] + 90 <= ptm->tm_year)
                    {
                        has_valid_year = 1;
                        printf("Made week %hd of %hd\n", edidchar[0x10], edidchar[0x11] + 1990);
                    }
                }
            }

            //Edid version part:
            printf("EDID version: %hd.%hd\n\n", edidchar[0x12], edidchar[0x13]);
            if (edidchar[0x12] == 1)
            {
                if (edidchar[0x13] > 4)
                {
                    printf("Claims > 1.4, assuming 1.4 conformance\n");
                    edidchar[0x13] = 4;
                }
                switch (edidchar[0x13])
                {
                    case 4:
                        claims_one_point_four = 1;
                    case 3:
                        claims_one_point_three = 1;
                    case 2:
                        claims_one_point_two = 1;
                    default:
                        break;
                }
                claims_one_point_oh = 1;
            }

            //Display part:
            if (edidchar[0x14] & 0x80)
            {
                int conformance_mask;
                analog = 0;
                printf("Digital display\n");
                if (claims_one_point_four)
                {
                    conformance_mask = 0;
                    if ((edidchar[0x14] & 0x70) == 0x00)
                        printf("Color depth is undefined\n");
                    else if ((edidchar[0x14] & 0x70) == 0x70)
                        nonconformant_digital_display = 1;
                    else
                        printf("%d bits per primary color channel\n",((edidchar[0x14] & 0x70) >> 3) + 4);

                    switch (edidchar[0x14] & 0x0f)
                    {
                        case 0x00: printf("Digital interface is not defined\n"); break;
                        case 0x01: printf("DVI interface\n"); break;
                        case 0x02: printf("HDMI-a interface\n"); break;
                        case 0x03: printf("HDMI-b interface\n"); break;
                        case 0x04: printf("MDDI interface\n"); break;
                        case 0x05: printf("DisplayPort interface\n"); break;
                        default:
                        nonconformant_digital_display = 1;
                    }
                }
                else if (claims_one_point_two)
                {
                    conformance_mask = 0x7E;
                    if (edidchar[0x14] & 0x01)
                    {
                        printf("DFP 1.x compatible TMDS\n");
                    }
                }
                else
                    conformance_mask = 0x7F;
                if (!nonconformant_digital_display)
                    nonconformant_digital_display = edidchar[0x14] & conformance_mask;
            }
            else
            {
                analog = 1;
                int voltage = (edidchar[0x14] & 0x60) >> 5;
                int sync = (edidchar[0x14] & 0x0F);
                printf("Analog display, Input voltage level: %s V\n",
                       voltage == 3 ? "0.7/0.7" :
                       voltage == 2 ? "1.0/0.4" :
                       voltage == 1 ? "0.714/0.286" :
                       "0.7/0.3");

                if (claims_one_point_four)
                {
                    if (edidchar[0x14] & 0x10)
                        printf("Blank-to-black setup/pedestal\n");
                    else
                        printf("Blank level equals black level\n");
                }
                else if (edidchar[0x14] & 0x10)
                {

                    //XXX this is just the X text.  1.3 says "if set, display expects
                    //a blank-to-black setup or pedestal per appropriate Signal
                    //Level Standard".  Whatever _that_ means.

                    printf("Configurable signal levels\n");
                }
                printf("Sync: %s%s%s%s\n", sync & 0x08 ? "Separate " : "",
                    sync & 0x04 ? "Composite " : "",
                    sync & 0x02 ? "SyncOnGreen " : "",
                    sync & 0x01 ? "Serration " : "");
            }

            if (edidchar[0x15] && edidchar[0x16])
                printf("Maximum image size: %d cm x %d cm\n", edidchar[0x15], edidchar[0x16]);
            else if (claims_one_point_four && (edidchar[0x15] || edidchar[0x16]))
            {
                if (edidchar[0x15])
                    printf("Aspect ratio is %f (landscape)\n", 100.0/(edidchar[0x16] + 99));
                else
                    printf("Aspect ratio is %f (portrait)\n", 100.0/(edidchar[0x15] + 99));
            }
            else
            {
                //Either or both can be zero for 1.3 and before
                printf("Image size is variable\n");
            }

            if (edidchar[0x17] == 0xff)
            {
                if (claims_one_point_four)
                    printf("Gamma is defined in an extension block\n");
                else
                    // XXX Technically 1.3 doesn't say this...
                    printf("Gamma: 1.0\n");
            }
            else
                printf("Gamma: %.2f\n", ((edidchar[0x17] + 100.0) / 100.0));

            if (edidchar[0x18] & 0xE0)
            {
                printf("DPMS levels:");
                if (edidchar[0x18] & 0x80)
                    printf(" Standby");
                if (edidchar[0x18] & 0x40)
                    printf(" Suspend");
                if (edidchar[0x18] & 0x20)
                    printf(" Off");
                printf("\n");
            }

            /* FIXME: this is from 1.4 spec, check earlier */
            if (analog)
            {
                switch (edidchar[0x18] & 0x18)
                {
                    case 0x00: printf("Monochrome or grayscale display\n"); break;
                    case 0x08: printf("RGB color display\n"); break;
                    case 0x10: printf("Non-RGB color display\n"); break;
                    case 0x18: printf("Undefined display color type\n");
                }
            }
            else
            {
                printf("Supported color formats: RGB 4:4:4");
                if (edidchar[0x18] & 0x10)
                    printf(", YCrCb 4:4:4");
                if (edidchar[0x18] & 0x08)
                    printf(", YCrCb 4:2:2");
                printf("\n");
            }

            if (edidchar[0x18] & 0x04)
                printf("Default (sRGB) color space is primary color space\n");
            if (edidchar[0x18] & 0x02)
            {
                printf("First detailed timing is preferred timing\n");
                has_preferred_timing = 1;
            }
            if (edidchar[0x18] & 0x01)
                printf("Supports GTF timings within operating range\n");

            /* XXX color section */

            printf("\nEstablished timings supported:\n");
            for (i = 0; i < 17; i++)
            {
                if (edidchar[0x23 + i / 8] & (1 << (7 - i % 8)))
                {
                    printf("  %dx%d@%dHz\n", established_timings[i].x, established_timings[i].y, established_timings[i].refresh);
                }
            }

            printf("\nStandard timings supported:\n");
            for (i = 0; i < 8; i++)
            {
                uint8_t b1 = edidchar[0x26 + i * 2], b2 = edidchar[0x26 + i * 2 + 1];
                unsigned int x, y, refresh;

                if (b1 == 0x01 && b2 == 0x01)
                    continue;

                if (b1 == 0)
                {
                    printf("non-conformant standard timing (0 horiz)\n");
                    continue;
                }
                x = (b1 + 31) * 8;
                switch ((b2 >> 6) & 0x3)
                {
                    case 0x00:
                        y = x * 10 / 16;
                        break;
                    case 0x01:
                        y = x * 3 / 4;
                        break;
                    case 0x02:
                        y = x * 4 / 5;
                        break;
                    case 0x03:
                        y = x * 9 / 15;
                        break;
                }
                refresh = 60 + (b2 & 0x3f);
                printf("  %dx%d@%dHz\n", x, y, refresh);
            }

            //Detailed timing part:
            has_valid_detailed_blocks = detailed_block(edidchar + 0x36, 0,&did_detailed_timing ,&claims_one_point_four);
            if (has_preferred_timing && !did_detailed_timing)
                has_preferred_timing = 0; // not really accurate...
            has_valid_detailed_blocks &= detailed_block(edidchar + 0x48, 0,&did_detailed_timing,&claims_one_point_four);
            has_valid_detailed_blocks &= detailed_block(edidchar + 0x5A, 0,&did_detailed_timing,&claims_one_point_four);
            has_valid_detailed_blocks &= detailed_block(edidchar + 0x6C, 0,&did_detailed_timing,&claims_one_point_four);
            printf("\n");
            return 0;
        }
        else
            printf("This wasn't really an EDID...\n");
    }
    else
    {
        printf("Insuficient parametres, application closed\n");
    }
    return 0;
}
